@@ -2,19 +2,32 @@
 {-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE MultiParamTypeClasses      #-}
+{-# LANGUAGE TypeFamilies               #-}
+{-# LANGUAGE UndecidableInstances       #-}
 
 module Snap.Snaplet.Internal.LensT where
 
-import           Control.Applicative
-import           Control.Category
-import           Control.Lens
-import           Control.Monad.CatchIO
-import           Control.Monad.Reader
-import           Control.Monad.State.Class
-import           Prelude hiding ((.), id, catch)
-import           Snap.Core
 
-import           Snap.Snaplet.Internal.RST
+------------------------------------------------------------------------------
+import           Control.Applicative         (Alternative (..),
+                                              Applicative (..))
+import           Control.Category            ((.))
+import           Control.Lens                (ALens', cloneLens, storing, (^#))
+import           Control.Monad               (MonadPlus (..))
+import           Control.Monad.Base          (MonadBase (..))
+import           Control.Monad.Reader        (MonadReader (..))
+import           Control.Monad.State.Class   (MonadState (..))
+import           Control.Monad.Trans         (MonadIO (..), MonadTrans (..))
+import           Control.Monad.Trans.Control (ComposeSt, MonadBaseControl (..),
+                                              MonadTransControl (..),
+                                              defaultLiftBaseWith,
+                                              defaultLiftWith, defaultRestoreM,
+                                              defaultRestoreT)
+import           Prelude                     (Functor (..), Monad (..), const,
+                                              ($), ($!))
+import           Snap.Core                   (MonadSnap (..))
+import           Snap.Snaplet.Internal.RST   (RST (..), runRST, withRST)
+------------------------------------------------------------------------------
 
 
 newtype LensT b v s m a = LensT (RST (ALens' b v) s m a)
@@ -24,10 +37,8 @@ newtype LensT b v s m a = LensT (RST (ALens' b v) s m a)
            , Applicative
            , MonadIO
            , MonadPlus
-           , MonadCatchIO
            , Alternative
-           , MonadReader (ALens' b v)
-           , MonadSnap )
+           , MonadReader (ALens' b v))
 
 
 ------------------------------------------------------------------------------
@@ -36,6 +47,30 @@ instance Monad m => MonadState v (LensT b v b m) where
     put = lPut
 
 
+instance MonadBase bs m => MonadBase bs (LensT b v s m) where
+    liftBase = lift . liftBase
+
+
+instance MonadBaseControl bs m => MonadBaseControl bs (LensT b v s m) where
+     type StM (LensT b v s m) a = ComposeSt (LensT b v s) m a
+     liftBaseWith = defaultLiftBaseWith
+     restoreM = defaultRestoreM
+     {-# INLINE liftBaseWith #-}
+     {-# INLINE restoreM #-}
+
+
+instance MonadTransControl (LensT b v s) where
+    type StT (LensT b v s) a = StT (RST (ALens' b v) s) a
+    liftWith = defaultLiftWith LensT (\(LensT rst) -> rst)
+    restoreT = defaultRestoreT LensT
+    {-# INLINE liftWith #-}
+    {-# INLINE restoreT #-}
+
+
+instance MonadSnap m => MonadSnap (LensT b v s m) where
+    liftSnap m = LensT $ liftSnap m
+
+
 ------------------------------------------------------------------------------
 getBase :: Monad m => LensT b v s m s
 getBase = LensT get
